#include "global.hh"
#include "Cfile.hh"
#include "Cconverter.hh"
#include "Cdata_reader.hh"

#include "header.hh"

Cconverter::Cconverter() : m_file_in(0), m_file_out(0)
{
  /* nothing */
}

Cconverter::~Cconverter()
{
  /* nothing */
}

int Cconverter::set_file(Cfile* file, std::string mode)
{
  if(mode == "in"){
    m_file_in = file;
  }else if(mode == "out"){
    m_file_out = file;
  }else{
    puts("Cconverter::set_file mode is wrong");
    return ERROR;
  }

  return OK;
}

int Cconverter::convert_binary_to_readable_binary()
{
  /* header block */
  {
    data_t header_block[BLOCK_SIZE/sizeof(data_t)];

    if(read_header_binary(header_block, m_file_in) == ERROR){
      printf("read_heaader_binary\n");
      return ERROR;
    }
    m_file_out->write_word('\n');
    write_header_readable_binary(header_block, m_file_out);
  }

  /* event block */
  {
    data_t event_block[BLOCK_SIZE/sizeof(data_t)];
    int i(0);
    /* read and write all event */
    while(1){
      /* check of ender block */
      m_file_in->read_one_block(event_block);
      m_file_in->seek_back_one_block();
      if((unsigned short)event_block[0] == FLAG_OF_ENDER){
	break;
      }
      if(read_event_block_binary(event_block, m_file_in) == ERROR){
	printf("read_event_block_binary\n");
	  return ERROR;
      }
      ++i;
      write_event_block_readable_binary(event_block, i, m_file_out);
    }
  }

  /* ender block */
  {
    data_t ender_block[BLOCK_SIZE/sizeof(data_t)];
    if(read_ender_binary(ender_block, m_file_in) == ERROR){
      printf("read_ender_binary\n");
      return ERROR;
    }
    write_ender_readable_binary(ender_block, m_file_out);
  }

  return OK;
}

int Cconverter::convert_binary_to_txt()
{
  /* header block */
  {
    data_t header_block[BLOCK_SIZE/sizeof(data_t)];
    if(read_header_binary(header_block, m_file_in) == ERROR){
      printf("read_heaader_binary\n");
      return ERROR;
    }
    m_file_out->write_word('\n');
    write_header_txt(header_block, m_file_out);
  }

  /* event block */
  {
    data_t event_block[BLOCK_SIZE/sizeof(data_t)];
    int i(0);
    /* read and write all event */
    while(1){
      /* check of ender block */
      m_file_in->read_one_block(event_block);
      m_file_in->seek_back_one_block();
      if((unsigned short)event_block[0] == FLAG_OF_ENDER){     
	break;
      }
      if(read_event_block_binary(event_block, m_file_in) == ERROR){
	printf("read_event_block_binary\n");
	return ERROR;
      }
      ++i;
      write_event_block_txt(event_block, i, m_file_out);
    }
  }
  /* ender block */
  {
    data_t ender_block[BLOCK_SIZE/sizeof(data_t)];
    if(read_ender_binary(ender_block, m_file_in) == ERROR){
      printf("read_ender_binary\n");
      return ERROR;
    }
    write_ender_txt(ender_block, m_file_out);
  }

  return OK;
}

int Cconverter::convert_txt_to_binary()
{
  /* header */
  {
    Cheader* header = new Cheader();
    read_header_txt(header);
    write_header_binary(header);
    delete header;
  }

  /* event block */
  {
    int event_num(0);
    while(1){
      int return_flag = write_event_block_binary(event_num);
      if(return_flag == END_OF_FILE){
	break;
      }else if(return_flag == ERROR){
	printf("return_flag\n");
	return ERROR;
      }
    }
  }

  /* ender */
  {
    Cender* ender = new Cender();
    read_ender_txt(ender);
    write_ender_binary(ender);
    delete ender;
  }

  return OK;
}

/* rdf to readable rdf or txt */
int Cconverter::read_header_binary(data_t header_block[], Cfile* file)
{
  file->read_one_block(header_block);
  if((unsigned short)header_block[0] != FLAG_OF_HEADER){
    puts("this block is not header");
    return ERROR;
  }
  return OK;
}

int Cconverter::read_event_block_binary(data_t event_block[], Cfile* file)
{
  file->read_one_block(event_block);
  if((unsigned short)event_block[0] != FLAG_OF_EVENT_BLOCK){
    puts("this block is not event block");
    return ERROR;
  }
  return OK;
}

int Cconverter::read_ender_binary(data_t ender_block[], Cfile* file)
{
  file->read_one_block(ender_block);
  if((unsigned short)ender_block[0] != FLAG_OF_ENDER){
    puts("this block is not ender");
    return ERROR;
  }
  return OK;
}

int Cconverter::write_header_readable_binary(data_t header_block[], Cfile* file)
{
  file->write_word("#  start header\n\n");
  file->write_one_block(header_block);
  file->write_word('\n');
  return OK;
}
int Cconverter::write_event_block_readable_binary(data_t event_block[], int i, Cfile* file)
{
  file->write_word("#  ");
  file->write_word(i);
  file->write_word(" event block\n\n");
  file->write_one_block(event_block);
  file->write_word('\n');

  return OK;
}
int Cconverter::write_ender_readable_binary(data_t ender_block[], Cfile* file)
{
  file->write_word("#  start ender\n\n");
  file->write_one_block(ender_block);
  file->write_word('\n');
  return OK;
}

int Cconverter::write_header_txt(data_t header_block[], Cfile* file)
{
  char buf[100];

  file->write_word("#  start header\n\n");

  /* run number */
  decode_binary_to_string(header_block, buf,
			  RUN_NUMBER_NUM, RUN_NUMBER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* start time */
  decode_binary_to_string(header_block, buf,
			  START_TIME_NUM, START_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* stop time */
  decode_binary_to_string(header_block, buf,
			  STOP_TIME_NUM, STOP_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* print time */
  decode_binary_to_string(header_block, buf,
			  PRINT_TIME_NUM, PRINT_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* print date */
  decode_binary_to_string(header_block, buf,
			  PRINT_DATE_NUM, PRINT_DATE_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* header */
  decode_binary_to_string(header_block, buf,
			  HEADER_NUM, HEADER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  /* ender */
  decode_binary_to_string(header_block, buf,
			  ENDER_NUM, ENDER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  file->write_word('\n');
  return OK;
}

int Cconverter::write_event_block_txt(data_t event_block[], int i, Cfile* file)
{
  char buf[100];
  char str[10];
  int data;
  int NOW_WORD_NUM(0);//現在のwordの位置

  file->write_word("#  ");
  file->write_word(i);
  file->write_word(" event block\n\n");

  NOW_WORD_NUM = FLAG_OF_EVENT_BLOCK_NUM;
  decode_binary_to_string(event_block, buf,
			  NOW_WORD_NUM,
			  FLAG_OF_EVENT_BLOCK_LEN);/* 空読み */
  NOW_WORD_NUM += FLAG_OF_EVENT_BLOCK_LEN/sizeof(data_t);
  decode_binary_to_string(event_block, buf,
			  NOW_WORD_NUM,
			  EVENT_BLOCK_BLANK_LEN);/* 空読み */
  NOW_WORD_NUM += EVENT_BLOCK_BLANK_LEN/sizeof(data_t);

  while(NOW_WORD_NUM < BLOCK_SIZE/sizeof(data_t) - 1){
    decode_binary_to_integer(event_block, &data,
			     NOW_WORD_NUM, EVENT_SIZE_LEN);
    NOW_WORD_NUM += EVENT_SIZE_LEN/sizeof(data_t);
    int event_size = data;
    if((event_size & 0x8000) == 0x0000){
      break;//error check
    }
    event_size &= 0x7fff;//先頭の1ビットを落とす

    decode_binary_to_integer(event_block, &data,
			     NOW_WORD_NUM, FID_LEN);
    NOW_WORD_NUM += FID_LEN/sizeof(data_t);
    int fid = data;
    if(fid != 1){
      break;//error check
    }

    decode_binary_to_integer(event_block, &data,
			     NOW_WORD_NUM, EVENT_ID_LEN);
    NOW_WORD_NUM += EVENT_ID_LEN/sizeof(data_t);
    int event_id = data;
    sprintf(str, "%4d ", event_id);
    file->write_word(str);

    decode_binary_to_integer(event_block, &data,
			     NOW_WORD_NUM, SEGMENT_SIZE_LEN);
    NOW_WORD_NUM += SEGMENT_SIZE_LEN/sizeof(data_t);
    int segment_size = data;

    /* error check */
    if((event_size - segment_size) !=
       (EVENT_SIZE - SEGMENT_SIZE)/sizeof(data_t)){
      printf("event_size or segment_size is wrong\n");
      return ERROR;
    }

    decode_binary_to_integer(event_block, &data,
			     NOW_WORD_NUM, SEGMENT_ID_LEN);
    NOW_WORD_NUM += SEGMENT_ID_LEN/sizeof(data_t);
    int segment_id = data;
    sprintf(str, "%4d ", segment_id);
    file->write_word(str);

    for(int j = 0; j < segment_size - SEGMENT_SIZE/sizeof(data_t); ++j){ 
      decode_binary_to_integer(event_block, &data,
			       NOW_WORD_NUM, DATA_LEN);
      NOW_WORD_NUM += DATA_LEN/sizeof(data_t) ;
      sprintf(str, "%4d ", (unsigned short)data);
      file->write_word(str);
    }
    file->write_word('\n');
  }
  file->write_word('\n');
  return OK;
}

int Cconverter::write_ender_txt(data_t ender_block[], Cfile* file)
{
  char buf[100];

  file->write_word("#  start ender\n\n");

  /* run number */
  decode_binary_to_string(ender_block, buf,
			  RUN_NUMBER_NUM, RUN_NUMBER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* start time */
  decode_binary_to_string(ender_block, buf,
			  START_TIME_NUM, START_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* stop time */
  decode_binary_to_string(ender_block, buf,
			  STOP_TIME_NUM, STOP_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* print time */
  decode_binary_to_string(ender_block, buf,
			  PRINT_TIME_NUM, PRINT_TIME_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* print date */
  decode_binary_to_string(ender_block, buf,
			  PRINT_DATE_NUM, PRINT_DATE_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* header */
  decode_binary_to_string(ender_block, buf,
			  HEADER_NUM, HEADER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");
  /* ender */
  decode_binary_to_string(ender_block, buf,
			  ENDER_NUM, ENDER_LEN);
  file->write_word("#  ");
  file->write_word(buf);
  file->write_word("\n");

  file->write_word('\n');
  return OK;
}

/* txt to rdf*/
int Cconverter::read_header_txt(Cuser* header)
{
  memcpy(header->m_run_number, user_run_number, RUN_NUMBER_LEN);
  memcpy(header->m_start_time, user_start_time, START_TIME_LEN);
  memcpy(header->m_stop_time, user_stop_time, STOP_TIME_LEN);
  memcpy(header->m_print_time, user_print_time, PRINT_TIME_LEN);
  memcpy(header->m_print_date, user_print_date, PRINT_DATE_LEN);

  memset(header->m_header, 0, HEADER_LEN);
  header->m_header_size = 0;
  while(user_header[header->m_header_size++] != '\0'){/* nothing */}
  memcpy(header->m_header, user_header, --header->m_header_size);

  memset(header->m_ender, 0, ENDER_LEN);
  header->m_ender_size = 0;
  while(user_ender[header->m_ender_size++] != '\0'){/* nothing */}
  memcpy(header->m_ender, user_ender, --header->m_ender_size);

  return OK;
}

int Cconverter::read_event_block_txt(char* buf, size_t size, Cfile* file)
{
  return file->read_line(buf, size);
}

int Cconverter::read_ender_txt(Cuser* ender)
{
  return read_header_txt(ender);
}

int Cconverter::write_header_binary(Cuser* header)
{
  char buf[BLOCK_SIZE];

  /* flag of header 1byte*/
  m_file_out->write_binary(&FLAG_OF_HEADER, sizeof(data_t));
  /* blank 9byte*/
  memset(buf, BLANK, HEADER_BLANK_LEN);
  m_file_out->write_binary(buf, HEADER_BLANK_LEN);
  /* run number 8byte*/
  m_file_out->write_binary(header->m_run_number, RUN_NUMBER_LEN);
  /* space 2byte*/
  m_file_out->write_binary(' ');
  m_file_out->write_binary(' ');
  /* start time 18byte*/
  m_file_out->write_binary(header->m_start_time, START_TIME_LEN);
  /* stop time 18byte*/
  m_file_out->write_binary(header->m_stop_time, STOP_TIME_LEN);    
  /* space 2yte*/
  m_file_out->write_binary(' ');
  m_file_out->write_binary(' ');
  /* print time 18byte*/
  m_file_out->write_binary(header->m_print_time, PRINT_TIME_LEN);    
  /* print date 10byte*/
  m_file_out->write_binary(header->m_print_date, PRINT_DATE_LEN);
  /* blank 4byte*/
  m_file_out->write_binary(&BLANK, sizeof(data_t));
  m_file_out->write_binary(&BLANK, sizeof(data_t));
  /* header 80byte*/
  m_file_out->write_binary(header->m_header, HEADER_LEN);
  /* ender 80byte*/
  m_file_out->write_binary(header->m_ender, ENDER_LEN);
  /* blank 0x4000 - 260 byte */
  memset(buf, 0, REST_LEN);
  m_file_out->write_binary(buf, REST_LEN);

  return OK;
}

int Cconverter::write_event_block_binary(int& event_num)
{
  char buf[BLOCK_SIZE];
  size_t byte_counter(0), last_byte(0), max_byte(0);
  int return_flag(0);
  int data_reader_flag(0);

  Cdata_reader* data_reader = new Cdata_reader();

  /* flag of event block 2byte*/
  m_file_out->write_binary(&FLAG_OF_EVENT_BLOCK, sizeof(data_t));
  byte_counter += FLAG_OF_EVENT_BLOCK_LEN;
  /* blank 6byte*/
  memset(buf, BLANK, EVENT_BLOCK_BLANK_LEN);
  m_file_out->write_binary(buf, EVENT_BLOCK_BLANK_LEN);
  byte_counter += EVENT_BLOCK_BLANK_LEN;

  /* initialize for event roop */
  int i(0);
  last_byte = byte_counter;
  /* read in until 16kbyte readen or file end */
  while(1){
    /* byte counting */
    if(max_byte < (byte_counter - last_byte)){
      max_byte = byte_counter - last_byte;
    }    
    last_byte = byte_counter;
    //    printf("%d %d %d\n", byte_counter, max_byte, byte_counter+max_byte);
    if(byte_counter + max_byte > BLOCK_SIZE){
      printf("\n goto next event block\n\n");
      return_flag = OK;
      break;
    }

    /* read in and check */
    data_reader_flag = data_reader->read_in_one_line(m_file_in);
    if(data_reader_flag == END_OF_FILE){
      printf("\n %d events converted \n", event_num);
      return_flag = END_OF_FILE;
      break;
    }else if(data_reader_flag == COMMENT){
      //      printf("comment\n");
      continue;
    }else if(data_reader_flag == ERROR){
      printf("data_reader_flag\n");
      return ERROR;
    }
    if(event_num%100 == 0){
      printf("now reading line #%d\n", event_num);	
    }
    ++event_num;
    ++i;

    /* main part */
    int number_of_word = data_reader->get_number_of_word();
    /* event size 2byte */
    {
      data_t temp = (number_of_word +
		     EVENT_SIZE/sizeof(data_t)) | FLAG_OF_EVENT_SIZE;
      m_file_out->write_binary(&temp, sizeof(data_t));
      byte_counter += EVENT_SIZE_LEN;
    }
    /* fid 2byte */
    m_file_out->write_binary(&FLAG_OF_FID, sizeof(data_t));
    byte_counter += FID_LEN;
    /* event id 2byte */
    /* you can use event_num(insted of i) for holl event num*/
    {
      data_t temp = i;
      m_file_out->write_binary(&temp, sizeof(data_t));
      byte_counter += EVENT_ID_LEN;
    }
    /* segment size 2byte */
    {
      data_t temp = number_of_word + SEGMENT_SIZE/sizeof(data_t);
      m_file_out->write_binary(&temp, sizeof(data_t));
      byte_counter += SEGMENT_SIZE_LEN;
    }
    /* segment id 2byte (what do this mean???) */
    {
      data_t temp = 0x0001;
      m_file_out->write_binary(&temp, sizeof(data_t));
      byte_counter += SEGMENT_ID_LEN;
    }
    /* data 2byte*number_of_word */
    for(int j = 0; j < number_of_word; ++j){
      data_t data;
      if(data_reader->get_data(data, j) == ERROR){
	printf("data_reader->get_data(%d, %d)\n", data, j);
	return ERROR;
      }
      m_file_out->write_binary(&data, sizeof(data_t));
      byte_counter += sizeof(data_t);
    }
  }
  /* write endof event block */
  m_file_out->write_binary(&FLAG_OF_END_OF_EVENT_BLOCK, sizeof(data_t));
  byte_counter += sizeof(data_t);
  m_file_out->write_binary(&FLAG_OF_END_OF_EVENT_BLOCK, sizeof(data_t));
  byte_counter += sizeof(data_t);

  /* write rest by blank */
  size_t rest_len(BLOCK_SIZE - byte_counter);
  memset(buf, 0, rest_len);
  m_file_out->write_binary(buf, rest_len);

  delete data_reader;
  
  return return_flag;
}

int Cconverter::write_ender_binary(Cuser* ender)
{
  char buf[BLOCK_SIZE];

  /* flag of ender 1byte*/
  m_file_out->write_binary(&FLAG_OF_ENDER, sizeof(data_t));
  /* blank 9byte*/
  memset(buf, BLANK, ENDER_BLANK_LEN);
  m_file_out->write_binary(buf, ENDER_BLANK_LEN);
  /* run number 8byte*/
  m_file_out->write_binary(ender->m_run_number, RUN_NUMBER_LEN);
  /* space 2byte*/
  m_file_out->write_binary(' ');
  m_file_out->write_binary(' ');
  /* start time 18byte*/
  m_file_out->write_binary(ender->m_start_time, START_TIME_LEN);
  /* stop time 18byte*/
  m_file_out->write_binary(ender->m_stop_time, STOP_TIME_LEN);    
  /* space 2yte*/
  m_file_out->write_binary(' ');
  m_file_out->write_binary(' ');
  /* print time 18byte*/
  m_file_out->write_binary(ender->m_print_time, PRINT_TIME_LEN);    
  /* print date 10byte*/
  m_file_out->write_binary(ender->m_print_date, PRINT_DATE_LEN);
  /* blank 4byte*/
  m_file_out->write_binary(&BLANK, sizeof(data_t));
  m_file_out->write_binary(&BLANK, sizeof(data_t));
  /* header 80byte*/
  m_file_out->write_binary(ender->m_header, HEADER_LEN);
  /* ender 80byte*/
  m_file_out->write_binary(ender->m_ender, ENDER_LEN);
  /* blank 0x4000 - 260 byte */
  memset(buf, 0, REST_LEN);
  m_file_out->write_binary(buf, REST_LEN);

  return OK;
}
